import collections
import operator


def calculate_distribution(values, accuracy=1, feature_distribution=None):
    """Calculate the distribution of the values.

    If `accuracy` == 1, this function simply counts the values and
    normalises their counts into probabilities.

    If `accuracy` < 1, then it mixes the counts with the
    `feature_distribution`, which you can think of as a prior for the
    feature. We assume that we draw from `values` `accuracy` of the time
    and from `feature_distribution` 1 - `accuracy` of the time.

    The case where `accuracy` < 1 is slow.
    """
    assert accuracy == 1 or feature_distribution is not None
    counts = collections.Counter(values)
    total_counts = sum(counts.values())
    if accuracy == 1:
        return {v: c / total_counts for v, c in counts.items()}
    else:
        assert counts.keys() <= feature_distribution.keys()
        acc_tc = accuracy / total_counts
        onemacc = 1 - accuracy
        probs = ((v, counts[v] * acc_tc + p * onemacc)
                 for v, p in feature_distribution.items())
        return {v: p for v, p in probs if p > 0}


def complete_feature_priors(df, feature_priors):
    feature_priors = feature_priors.copy()
    for i in range(len(df.columns)):
        if i in feature_priors:
            fd = feature_priors[i]
            #check if complete?
        else:
            ffs = tuple(map(operator.itemgetter(i), df.values))
            fd = calculate_distribution(ffs)
        feature_priors[i] = fd
    return feature_priors


import pkg_resources

from piflib.pif_calculator import compute_cigs, compute_weighted_cigs, compute_csfs

try:
    __version__ = pkg_resources.get_distribution('clkhash').version
except pkg_resources.DistributionNotFound:
    __version__ = "development"

__author__ = "Data61"




import collections
import math
import operator
import random
import itertools


import pandas as pd
import numpy as np

from piflib.data_util import calculate_distribution, complete_feature_priors
from piflib.entropy import create_conditional_entropy_table


def compute_cigs(dataframe,
                 feature_priors={},
                 feature_accuracies={},
                 samples=None):
    """Compute the cell information gain (CIG) for all cells in the dataset.

    Find the risk (as KL divergence from prior) for all attributes.

    :param dataframe: a Pandas DataFrame object containing tabular data
    :param feature_priors: feature_priors are optional. It is a dictionary mapping the
        feature index to an assumed prior. If not provided, the prior for
        the feature is calculated from the global distribution.
    :param feature_accuracies: `feature_accuracies` maps the feature index to the accuracy of the
        feature. If not provided for a feature, it defaults to 1.
    :return: a Pandas DataFrame containing the CIG values. The CIG values are at the same index as their corresponding
        cell values in the input dataframe.
    """
    unknown_features = 1
    dataset = dataframe.values
    l = len(dataset[0])
    assert all(len(row) == l for row in dataset)

    feature_priors = complete_feature_priors(dataframe, feature_priors)

    feature_accuracies = feature_accuracies.copy()
    for i in range(l):
        if i not in feature_accuracies:
            feature_accuracies[i] = 1

    feature_counts = [0] * l
    feature_kls = [[0] * len(dataset) for _ in range(l)]
    for is_ in sample_is(l, unknown_features, samples):
        feature_kls_this = find_kls_for_features(
            dataset,
            is_,
            feature_priors,
            feature_accuracies)
        for i, feature_kl in zip(is_, feature_kls_this):
            feature_kl_previous = feature_kls[i]
            feature_kls[i] = tuple(map(
                operator.add, feature_kl, feature_kl_previous))

        for i in is_:
            feature_counts[i] += 1

    for i, denom in enumerate(feature_counts):
        feature_kls[i] = tuple(map(
            operator.truediv,
            feature_kls[i],
            itertools.repeat(denom)))

    return pd.DataFrame(list(zip(*feature_kls)), columns=dataframe.columns)


def compute_weighted_cigs(dataframe, feature_priors={}, feature_accuracies={}):
    """Compute the Weighted Cell Information Gain (wCIG) for all cells in the dataset.

    Find the risk (as KL divergence from prior) for all attributes.

    :param dataframe: a Pandas DataFrame object containing tabular data
    :param feature_priors: feature_priors are optional. It is a dictionary mapping the
        feature index to an assumed prior. If not provided, the prior for
        the feature is calculated from the global distribution.
    :param feature_accuracies: `feature_accuracies` maps the feature index to the accuracy of the
        feature. If not provided for a feature, it defaults to 1.
    :return: a Pandas DataFrame containing the wCIG values. The wCIG values are at the same index as their
        corresponding cell values in the input dataframe.
        """
    cigs = compute_cigs(dataframe, feature_priors=feature_priors, feature_accuracies=feature_accuracies)
    cond_entropy = create_conditional_entropy_table(dataframe)
    weights = cond_entropy['H(X|Y)'].values / cond_entropy['H(X)']
    weights = np.nan_to_num(weights)
    return (cigs * weights).round(2)


def compute_csfs(df, feature_priors={}, feature_accuracies={}):
    """Compute the Cell Surprise Factor (CSF) for all cells in the dataset.

    The CSF id defined as the change in probability for a cell value between the prior and the posterior distribution.

    :param dataframe: a Pandas DataFrame object containing tabular data
    :param feature_priors: feature_priors are optional. It is a dictionary mapping the
        feature index to an assumed prior. If not provided, the prior for
        the feature is calculated from the global distribution.
    :param feature_accuracies: `feature_accuracies` maps the feature index to the accuracy of the
        feature. If not provided for a feature, it defaults to 1.
    :return: a Pandas DataFrame containing the CSF values. The CSF values are at the same index as their
        corresponding cell values in the input dataframe.
    """
    dataset = df.values
    l = len(dataset[0])
    # compute priors
    feature_priors = complete_feature_priors(df, feature_priors)
    feature_accuracies = feature_accuracies.copy()
    for i in range(l):
        if i not in feature_accuracies:
            feature_accuracies[i] = 1
    feature_csfs = [[0] * len(dataset) for _ in range(l)]
    for is_ in sample_is(l, 1, None):
        feature_csfs[is_[0]] = apply_to_posterior_and_prior(
            dataset,
            is_,
            feature_priors,
            feature_accuracies,
            calculate_prob_change)

    return pd.DataFrame(list(zip(*feature_csfs)), columns=df.columns)


def compute_posterior_distributions(feature, df):
    known_features = tuple(col_name for col_name in df.columns if col_name != feature)
    bucket = collections.defaultdict(list)
    bucket_map = []
    for idx, row in df.iterrows():
        key = tuple(row[known_feature] for known_feature in known_features)
        bucket[key].append(row[feature])
        bucket_map.append(key)
    bucket_distributions = {key: calculate_distribution(el_bucket) for key, el_bucket in bucket.items()}
    feature_vals = df[feature].unique()
    dists = {}
    for key, distribution in bucket_distributions.items():
        dists[str(key)] = [distribution.get(feature_val, 0) for feature_val in feature_vals]
    return dists, feature_vals


def binom(n, r):
    """ return binomial coefficient: n choose k"""
    return math.factorial(n) // math.factorial(n - r) // math.factorial(r)


def sample_is(n, r, samples):
    if samples is None:
        yield from itertools.combinations(range(n), r)
    else:
        total_combinations = binom(n, r)
        if samples > total_combinations:
            raise ValueError('more samples than combinations')
        if samples >= total_combinations >> 1:
            all_combinations = list(itertools.combinations(range(n), r))
            random.shuffle(all_combinations)

            num_produced = 0
            feature_produced = [False] * n
            for i, comb in enumerate(all_combinations):
                if num_produced >= samples:
                    break
                if all(map(feature_produced.__getitem__, comb)):
                    continue
                for j in comb:
                    feature_produced[j] = True
                num_produced += 1
                all_combinations[i] = None
                yield comb

            for comb in all_combinations:
                if num_produced >= samples:
                    break
                if comb is not None:
                    yield comb
        else:
            already_produced = set()
            feature_produced = [False] * n
            while len(already_produced) < samples:
                comb = random.sample(range(n), r)
                comb = tuple(sorted(comb))
                if (comb not in already_produced
                        and (all(already_produced)
                             or not all(map(already_produced.__getitem__,
                                            comb)))):
                    already_produced.add(comb)
                    for i in comb:
                        feature_produced[i] = True
                    yield comb


def apply_to_posterior_and_prior(dataset, feature_idx, prior_distributions, accuracies, fun):
    l = len(dataset[0])
    assert all(len(row) == l for row in dataset)
    feature_idx = feature_idx[0]
    buckets = collections.defaultdict(list)
    bucket_map = []
    for row in dataset:
        key = tuple(row[i] for i in range(l) if not i == feature_idx)
        buckets[key].append(row[feature_idx])
        bucket_map.append((key, row[feature_idx]))
    bucket_values = {
            key: fun(
                calculate_distribution(bucket,
                                       accuracy=accuracies[feature_idx],
                                       feature_distribution=prior_distributions[feature_idx]),
                prior_distributions[feature_idx])
            for key, bucket in buckets.items()}
    return [bucket_values[post_key][val] for post_key, val in bucket_map]


def find_kls_for_features(dataset, feature_is, feature_distributions, accuracies):
    """Find the KL divergence of feature values against the prior.

    We find the true distribution of the features taking into account
    the accuracy. We then compute the KL divergence.
    """
    l = len(dataset[0])
    assert all(len(row) == l for row in dataset)

    # one bucket per set of 'known' features
    buckets = [collections.defaultdict(list) for _ in range(len(feature_is))]
    bucket_map = [[] for _ in range(len(feature_is))]  ########
    for row in dataset:
        key = tuple(row[i] for i in range(l) if i not in feature_is)
        for i, j in enumerate(feature_is):
            buckets[i][key].append(row[j])
            bucket_map[i].append(key)

    bucket_kls = [
        {
            key: calculate_kl(
                calculate_distribution(bucket,
                                       accuracy=accuracies[feature_is[i]],
                                       feature_distribution=feature_distributions[feature_is[i]]),
                feature_distributions[feature_is[i]])
            for key, bucket in feature_buckets.items()}
        for i, feature_buckets in enumerate(buckets)]

    return [tuple(map(bucket_kls[i].__getitem__, bucket_map[i]))
            for i in range(len(feature_is))]


def calculate_kl(p, q):
    """Calculate D_KL(P || Q) (the KL-divergence) in bits.

    D_KL(P || Q) is the `information gained when one revises one's
    beliefs from the prior probability distribution Q to the posterior
    probability distribution P`. (Wikipedia, Kullback–Leibler
    divergence)

    `p` and `q` are both dictionaries mapping some hashable to a number.
    It is assumed that they are both normalised: their values should add
    up to 0. `q` must not have any 0 values unless the corresponding `p`
    value is also 0.
    """
    return sum(pk * math.log2(pk / q[k])
               for k, pk in p.items()
               if pk > 0)


def calculate_prob_change(p, q):
    """ calculate the change in probability for each element of the posterior compared to the prior"""
    return {k: abs(v - q[k]) for k, v in p.items()}



import setuptools

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

requirements = [
        "numpy>=1.19.0",
        "pandas>=1.1.0"
    ]

setuptools.setup(
    name="piflib",
    version="0.1.0",
    author="CSIRO's DATA61",
    author_email="confidential.computing@data61.csiro.au",
    description="A library for computing the personal information factor (PIF)",
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/PIFtools/piflib',
    license='Apache',
    install_requires=requirements,
    packages=setuptools.find_packages(),
    project_urls={
        'Documentation': 'http://piflib.readthedocs.io/',
        'Source': 'https://github.com/PIFtools/piflib',
        'Tracker': 'https://github.com/PIFtools/piflib/issues',
        'Discussion': 'https://github.com/PIFtools/piflib/discussions'
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    tests_require=[
        "pytest>=5.0",
    ]
)



"""Implement various entropy measures to improve CIG, RIG."""
import itertools
import numpy as np
import pandas as pd


def get_marginal(df, columns):
    """Return joint distribution of given columns.

    :param df: DataFrame that contains all observations
    :param columns: list of features to get joint distribution from

    :return dist: dict of outcome and corresponding probability
    """
    feature = list(df.columns)[0]
    d = pd.DataFrame(df.groupby(columns)[feature].count() / len(df))
    dist = d[feature].to_dict()
    return dist


def entropy(df, rvs):
    """Compute entropy of given random variables.

    :param df: Dataframe that contains all data
    :param rvs: list of feature names

    :return H: entropy of rvs
    """
    dist = get_marginal(df, rvs)
    ps = np.array(list(dist.values()))
    H = -sum(ps * np.log2(ps))
    return H


def mutual_information(df, rvs_X, rvs_Y):
    """Compute mutual information of two sets of random variables - I(X; Y).

    :param df: Dataframe that contains all data
    :param rvs_X: list of feature names for set X
    :param rvs_Y: list of feature names for set Y

    :return I: mutual information
    """
    H_X = entropy(df, rvs_X)
    H_Y = entropy(df, rvs_Y)
    H_XY = entropy(df, rvs_X + rvs_Y)
    I = H_X + H_Y - H_XY
    return(I)


def pairwise_mutual_information(df):
    """Compute the mutual information between all pairs of features in the dataset.

    Returns 3 lists, names of first feature, names of second feature, mutual information between the two.

    :param df: Dataframe that contains all data

    :return x: list of names of first feature
    :return y: list of names of second feature
    :return mi_xy: list of mutual information between every pair of feature names
    """
    x = []
    y = []
    mi_xy = []
    for feature_1, feature_2 in itertools.combinations(df.columns, 2):
        x.append(feature_1)
        y.append(feature_2)
        mi_xy.append(mutual_information(df, [feature_1], [feature_2]))
    return x, y, mi_xy


def create_entropy_mi_table(df, round_digits=2):
    """Create a dataframe containing the individual features' entropies and the pairwise mutual information.

    :param df: Dataframe that contains all data
    :round_digits: integer that specifies the number of decimal to disply

    :return emi_df: DataFrame that contains all pairs of mutual information
    """
    x, y, mi = pairwise_mutual_information(df)
    entropies = {name: entropy(df, [name]) for name in df}
    hx = [entropies[i] for i in x]
    hy = [entropies[i] for i in y]
    emi_df = pd.DataFrame({'X': x, 'Y': y, 'H(X)': hx, 'H(Y)': hy, 'I(X;Y)': mi}).round(round_digits)
    return emi_df


def conditional_entropy(df, rvs_X, rvs_Y):
    """Compute conditional entropy of set X given set Y - H(X|Y).

    :param df: Dataframe that contains all data
    :param rvs_X: list of feature names for set X
    :param rvs_Y: list of feature names for set Y

    :return H_XgY: conditional entropy
    """
    MI_XY = mutual_information(df, rvs_X, rvs_Y)
    H_X = entropy(df, rvs_X)
    H_XgY = H_X - MI_XY
    return H_XgY


def create_conditional_entropy_table(df, round_digits=2):
    """Create a dataframe containing the individual features' entropies and the conditional entropy
    given the rest of all features.

    :param df: Dataframe that contains all data
    :round_digits: integer that specifies the number of decimal to disply

    :return cond_df: DataFrame that contains conditional entropy of every feature given the rest of features
    """
    x, y, hx, hy, co_en = [], [], [], [], []
    for feature in df.columns:
        f_ys = list(df.columns.drop(feature))
        x.append(feature)
        y.append(f_ys)
        hx.append(entropy(df, [feature]))
        hy.append(entropy(df, f_ys))
        co_en.append(conditional_entropy(df, [feature], f_ys))
    cond_df = pd.DataFrame({'X': x, 'Y': y, 'H(X)': hx, 'H(Y)': hy, 'H(X|Y)': co_en}).round(round_digits)
    return cond_df


def dual_total_correlation(df, rvs):
    """Compute dual total correlation of given rvs.

    It is the amount of information that is shared among the variables.

    :param df: Dataframe that contains all data
    :param rvs: list of feature names

    :return tc: total correlation
    """
    others = lambda rv, rvs: list(set(rvs)- set(rv))
    sum_cond_entropy = sum(conditional_entropy(df, [rv], others([rv], rvs)) for rv in rvs)
    joint_entropy = entropy(df, rvs)
    tc = joint_entropy - sum_cond_entropy
    return tc


def residual_entropy(df, rvs):
    """Compute dual total correlation of given rvs.

    It is dual in the sense that together they form the entropy of the distribution.

    :param df: Dataframe that contains all data
    :param rvs: list of feature names

    :return sum_cond_entropy: residual entropy
    """
    others = lambda rv, rvs:list(set(rvs)- set(rv))
    sum_cond_entropy = sum(conditional_entropy(df, [rv], others([rv], rvs)) for rv in rvs)
    return sum_cond_entropy
